#------------------------------------------------------------------
# csvio.awk --- facilities for handling CSV data (gawk version)
#
# Author: Manuel Collado, <m-collado@users.sourceforge.net>, http://mcollado.z15.es
# License: Public domain - see UNLICENSE for details
# Version: 0.0 - Mar 2023
#
# Public functions:
#
#   csvimport([<comma> [, <quote>]])   (single chars)
#       Import a CSV record as a regular awk record delimited by OFS.
#       <comma>   Optional: The input CSV field delimiter (default comma ',')
#       <quote>   Optional: The input CSV quote char (default '"')
#
#   csvrecord(<afield> [, <comma> [, <quote>]])
#       Returns a CSV formatted record from a collection of fields.
#       <afield>  Array of field values indexed 1..N, or null to convert $0
#       <comma>   Optional: The output CSV field delimiter (default comma ',')
#       <quote>   Optional: The output CSV quote char (default '"')
#
#   csvrecord0(<comma> [, <quote>])
#       Just for convenience. Better use csvrecord().
#       Equivalent to csvrecord("", <comma>, <quote>)
#       For compatibility with the posix awk version.
#
# Local mode variables:
#   _csv_fpat[]: cached FPAT regexes to parse CSV fields
#------------------------------------------------------------------

#------------------------------------------------------------------
# Import a CSV record. First line already read.
#------------------------------------------------------------------

function csvimport(comma, quote,        cm, qt, more, af, k, fk, ofs){
    # Process arguments
    cm = comma ? comma : ","
    qt = quote ? quote : "\""
    if (!((cm qt) in _csv_fpat)) {
        # Build and store a suitable FPAT
        # CSV FPAT: "([^,]*)|(\"([^\"]|\"\")+\"[^,]*)"
        _csv_fpat[cm qt] = "([^" cm "]*)|(" qt "([^" qt "]|" qt qt ")+" qt "[^" cm "]*)"
    }
    # Collect multi-line data, if it is the case
    # If neccesary, add more input lines to ensure an even number of quotes
    while (gsub(qt, qt, $0) % 2 == 1 && (getline more) > 0) {
        $0 = $0 "\n" more
        NR--
        FNR--
    }
    # Strip unwanted CRs
    gsub("\r", "", $0)
    # Decode the CSV fields and rebuild $0
    patsplit($0, af, _csv_fpat[cm qt])
    $0 = ""     # Start rebuilding $0
    ofs = ""    # No delimiter before the first field
    for (k=1; k in af; k++) {
        fk = af[k]
        if (fk ~ ("^" qt ".*" qt "$")) {
            fk = substr(fk, 2, length(fk)-2)
            gsub(qt qt, qt, fk)
        }
        $0 = $0 ofs fk   # Concatenate fields, delimited by OFS
        ofs = OFS
    }
}

#------------------------------------------------------------------
# Generate a CSV record from a collection of fields
#------------------------------------------------------------------

# Compose a CSV record from an array of fields or from $0
function csvrecord(af, comma, quote,        r0, nf, record, k, sep, field) {
    if (isarray(af)) {
        r0 = 0
        nf = length(af)
    } else if (!af) {
        r0 = 1
        nf = NF
    } else {    # invalid argument
        return ""
    }
    comma = comma ? comma : ","
    quote = quote ? quote : "\""
    record = ""
    sep = ""
    for (k=1; k<=nf; k++) {
        field = r0 ? $k : af[k]
        if (index(field, comma) || index(field, quote) || index(field, "\n")) {
            gsub(quote, quote quote, field)
            field = quote field quote
        }
        record = record sep field
        sep = comma
    }
    return record
}

# Compose a CSV record from $0
function csvrecord0(comma, quote) {
    return csvrecord("", comma, quote)
}

